@@ -102,7 +102,7 @@ class AbstractSelectable:
         return (results)
 
 
-    def select(self, query, flattening_of_results=False):
+    def select(self, query, flattening_of_results=False, debug_mode = False):
         """given SELECT query returns Python list"""
         """Columns give the number of selected columns"""
         
@@ -113,7 +113,8 @@ class AbstractSelectable:
             return(self) #to enable chaining
         
         else:
-            print(query)
+            if debug_mode:
+                print(query)
             self.db1.cursor.execute(query)
             
             columns_count=len(self._get_selected_columns(query))
@@ -131,17 +132,18 @@ class AbstractSelectable:
             return(rows)
             
 
-    def select_all(self):
+    def select_all(self, debug_mode = False):
+        quote = self.db1.identifier_quote
         all_cols_query = ""
         for col in self.columns:
-            all_cols_query = all_cols_query + col + ","
+            all_cols_query = all_cols_query + quote + col + quote + ","
         if all_cols_query[-1] == ",":
             all_cols_query = all_cols_query[:-1]
-        list1 = self.select(f"SELECT {all_cols_query} FROM " + self.name)
+        list1 = self.select(f"SELECT {all_cols_query} FROM {quote}{self.name}{quote};", debug_mode = debug_mode)
         return (list1)
 
-    def select_to_df(self):
-        rows = self.select_all()
+    def select_to_df(self, debug_mode = False):
+        rows = self.select_all(debug_mode = debug_mode)
         if self.query_building_enabled:
             self.to_df()
             df=None
@@ -241,14 +243,9 @@ class AbstractTable(AbstractJoinable, abc.ABC):
         if self.columns is not None and self.types is not None:
             assert len(self.columns) == len(self.types)
             
-            # Hotfix: flatten nested lists (otherwise crashes might happen)
-            # TODO: Search for the causes and implement better handling
-            for i, column in enumerate(self.columns):
-                self.columns[i] = column[0] if type(column) == list else column
             self.column_type_dict={self.columns[i]:self.types[i] for i,x in enumerate(self.columns)}
         else:
             self.column_type_dict={}
-        
 
     # Temporary disabled, please make sure this is implemented where needed, don't introduce breaking changes please
     # @abc.abstractmethod
@@ -261,21 +258,23 @@ class AbstractTable(AbstractJoinable, abc.ABC):
         return cls(db1, name, columns, types, id_column_name=id_column_name)
     
     
-    def drop(self):
+    def drop(self, debug_mode = False):
         query = "DROP TABLE " + self.name
-        print(query)
+        if debug_mode:
+            print(query)
         self.execute(query)
 
-    def update(self, variable_assign, where=None):
+    def update(self, variable_assign, where=None, debug_mode = False):
         if where is None:
             query = "UPDATE " + self.name + " SET " + variable_assign
         else:
             query = "UPDATE " + self.name + " SET " + variable_assign + " WHERE " + where
-        print(query)
+        if debug_mode:
+            print(query)
         return self.execute(query)
 
     def update_from_df(
-            self, update_df: pd.DataFrame, where_column: Optional[str] = None, where_value: Any = None) -> None:
+            self, update_df: pd.DataFrame, where_column: Optional[str] = None, where_value: Any = None, debug_mode = False) -> None:
         """Build UPDATE SQL query from a dataframe and execute it.
 
         :param update_df: Dataframe with updated values - MUST only hold a single row
@@ -288,35 +287,45 @@ class AbstractTable(AbstractJoinable, abc.ABC):
         if not len(update_df) == 1:
             raise ValueError("There can only be one row in the UPDATE dataframe")
 
-        types_without_id_column = [type_ for column, type_ in zip(self.columns, self.types) 
-                           if column != self.id_column_name]
+        types_without_id_column = [
+            (column, type_) for column, type_ in zip(self.columns, self.types)
+            if column != self.id_column_name
+        ]
+
         if len(types_without_id_column) != len(update_df.columns):
             raise AttributeError(
                 "Number of columns in dataframe does not match number of columns in table"
             )
 
+        # Ensure the same order of columns in DF as in DB table
+        update_df_row_list = [
+            (column_name, update_df.iloc[0][column_name], column_type)
+            for column_name, column_type in types_without_id_column
+        ]
+
         column_value_string = ""
-        for (column, cell_value), column_type in zip(update_df.iloc[0].items(), types_without_id_column):
+        for column_name, cell_value, column_type in update_df_row_list:
             if cell_value is None:
-                column_value_string += f"{column} = NULL, "
+                column_value_string += f"{column_name} = NULL, "
             elif column_type in ["double", "int", "tinyint"]:
-                column_value_string += f"{column} = {cell_value}, "
+                column_value_string += f"{column_name} = {cell_value}, "
             elif "varchar" in column_type:
-                column_value_string += f"{column} = '{cell_value}', "
+                column_value_string += f"{column_name} = '{cell_value}', "
             elif column_type in ["json", "text", "mediumtext", "longtext", "datetime"]:
-                column_value_string += f"{column} = '{cell_value}', "
+                column_value_string += f"{column_name} = '{cell_value}', "
             else:
                 raise AttributeError(f"Unknown column type '{column_type}'")
 
         column_value_string = column_value_string.rstrip(", ")
-        sql_query = f"UPDATE {self.name} SET {column_value_string}"
+        quote = self.db1.identifier_quote
+        sql_query = f"UPDATE {quote}{self.name}{quote} SET {column_value_string}"
 
         if where_column is not None and where_value is not None:
             sql_query += f" WHERE {where_column} = {where_value};"
         else:
             sql_query += ";"
-
-        print(sql_query)
+        if debug_mode:
+            print(sql_query)
         self.execute(sql_query)
 
     def _adjust_df(self, df: pd.DataFrame, debug_mode=False) -> pd.DataFrame:
@@ -352,6 +361,8 @@ class AbstractTable(AbstractJoinable, abc.ABC):
 
         return df_copy
 
+    def extract_last_id(self) -> Any:
+        raise NotImplementedError("Method not implemented for this subclass")
 
     def insert_from_df(self, df, batch=1, try_mode=False, debug_mode=False, adjust_df=False, insert_id=False):
         if debug_mode:
@@ -398,7 +409,9 @@ class AbstractTable(AbstractJoinable, abc.ABC):
         #             rows[i][j] = "'" + record + "'"
         #print(rows)
         rows = df.values.tolist()
-        return self.insert(rows, batch=batch, try_mode=try_mode, debug_mode=False, insert_id=insert_id)
+        result = self.insert(rows, batch=batch, try_mode=try_mode, debug_mode=False, insert_id=insert_id)
+        self.db1.last_table_inserted_into = self.name
+        return result
 
     #TODO: need to solve inserting in different column_order
     #check df column names, permute if needed
@@ -412,11 +425,12 @@ class AbstractTable(AbstractJoinable, abc.ABC):
 
 
     def delete(self, where=None):
+        quote = self.db1.identifier_quote
 
         if where is None:
-            query = "DELETE FROM " + self.name
+            query = "DELETE FROM {quote}{self.name}{quote}"
         else:
-            query = "DELETE FROM " + self.name + " WHERE " + where
+            query = f"DELETE FROM {quote}{self.name}{quote} WHERE {where}"
         return self.execute(query)
 
 
