@@ -2,7 +2,7 @@ import pandas as pd
 import numpy as np
 from typing import Optional, Any
 import abc
-
+import time
 #xlsx imports
 import pathlib
 
@@ -83,20 +83,21 @@ class PostgresTable(AbstractTable):
         print("==========================================")
 
     def initialize_columns(self):
-        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
-        query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "';"
-        columns = information_schema_table.select(query)
+        """
+        TODO Dominik: Check for usecases of this method. Isn't it somewhat duplicated by `get_all_columns`?
+        """
+        columns = self.get_all_columns()
         self.columns = columns
 
     def initialize_types(self):
         self.types = self.get_all_types()
 
     def get_all_columns(self):
-        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
-        query = "SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "'"
-        columns = information_schema_table.select(query)
+        information_schema_table = PostgresTable(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
+        query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '{self.name}';"
+        columns = information_schema_table.select(query, flattening_of_results=True)
 
-        return (columns)
+        return columns
 
     def convert_types_from_mysql(self):
         inverse_dict_mysql_to_postgres = dict(zip(POSTGRES_TO_MYSQL_DATA_MAPPING.values(), POSTGRES_TO_MYSQL_DATA_MAPPING.keys()))
@@ -104,8 +105,7 @@ class PostgresTable(AbstractTable):
         self.types = postgres_types
 
     def get_all_types(self):
-
-        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
+        information_schema_table = PostgresTable(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
         query = "SELECT DATA_TYPE,character_maximum_length FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "'"
         types = information_schema_table.select(query)
         data_types = [x[0].lower() for x in types]
@@ -141,7 +141,7 @@ class PostgresTable(AbstractTable):
         return (cls(db1, name, columns, types))
 
     # @save_migration
-    def create(self, foreign_keys=None):
+    def create(self, foreign_keys=None, debug_mode = False):
         assert len(self.columns) == len(self.types)
         assert self.columns[0] == self.id_column_name
         assert self.types[0].lower() == "int" or self.types[0].lower() == "integer"
@@ -151,7 +151,8 @@ class PostgresTable(AbstractTable):
 
         query = query[:-1]
         query += ");"
-        print(query)
+        if debug_mode:
+            print(query)
         try:
             self.db1.execute(query)
         except Exception as e:
@@ -484,18 +485,18 @@ class SqlServerTable(AbstractTable):
         return (cls(db1, name, columns, types))
 
     def get_all_columns(self):
-        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
+        information_schema_table = SqlServerTable(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
         query = "SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "'"
-        columns = information_schema_table.select(query)
+        columns = information_schema_table.select(query, flattening_of_results=True)
         return (columns)
 
     def get_all_types(self):
-        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
+        information_schema_table = SqlServerTable(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
         query = "SELECT DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "'"
         types = information_schema_table.select(query)
         return (types)
 
-    def create(self):
+    def create(self, debug_mode = False):
         assert len(self.columns) == len(self.types)
         assert self.columns[0] == self.id_column_name
         assert self.types[0].lower() == "int"
@@ -503,8 +504,8 @@ class SqlServerTable(AbstractTable):
         for i in range(1, len(self.columns)):
             query += self.columns[i] + " " + self.types[i] + ","
         query += "PRIMARY KEY("+self.id_column_name+"))"
-
-        print(query)
+        if debug_mode:
+            print(query)
         try:
             self.db1.execute(query)
         except Exception as e:
@@ -620,9 +621,10 @@ class MysqlTable(AbstractTable):
     #     return cls(db1, name, columns, types, id_column_name=id_column_name)
 
     def initialize_columns(self):
-        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
-        query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{self.db1.DB_DATABASE}' AND  TABLE_NAME  = '" + self.name + "';"
-        columns = information_schema_table.select(query)
+        """
+        TODO Dominik: Check for usecases of this method. Isn't it somewhat duplicated by `get_all_columns`?
+        """
+        columns = self.get_all_columns()
         self.columns = columns
 
     def convert_types_from_mysql(self):
@@ -632,14 +634,13 @@ class MysqlTable(AbstractTable):
         self.types = self.get_all_types()
 
     def get_all_columns(self):
-        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
-        query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{self.db1.DB_DATABASE}' AND TABLE_NAME  = '" + self.name + "'"
-        columns = information_schema_table.select(query)
+        information_schema_table = MysqlTable(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
+        query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{self.db1.DB_DATABASE}' AND TABLE_NAME  = '{self.name}';"
+        columns = information_schema_table.select(query, flattening_of_results=True)
 
-        return (columns)
+        return columns
 
     def get_all_types(self):
-
         data_types, data_lengths = self.get_data_types_and_character_lengths()
         for i in range(len(data_types)):
             if data_lengths[i] is not None:
@@ -654,8 +655,8 @@ class MysqlTable(AbstractTable):
         'varchar' in the data_types list and 2047 in the data_lengths list.
     """
     def get_data_types_and_character_lengths(self):
-        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
-        query = f"SELECT DATA_TYPE,character_maximum_length FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{self.db1.DB_DATABASE}' AND TABLE_NAME  = '" + self.name + "'"
+        information_schema_table = MysqlTable(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
+        query = f"SELECT DATA_TYPE,character_maximum_length FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{self.db1.DB_DATABASE}' AND TABLE_NAME  = '{self.name}'"
         types = information_schema_table.select(query)
         data_types = [x[0] for x in types]
         data_lengths = [x[1] for x in types]
@@ -675,8 +676,20 @@ class MysqlTable(AbstractTable):
 
         return python_types
 
+    def extract_last_id(self) -> Any:
+        """
+        Extract the last inserted ID from the DB.
+
+        LAST_INSERT_ID exists in the DB connection context, therefore is safe to use if DB session is request-scoped
+        In this case we only use global connection, but we use Lock to ensure thread-safety across different requests
+        This is a go-to mechanism for extracting the ID of the inserted record for multiple SQL DBs,
+        altho this specific query is applicable only to MySQL.
+        """
+        assert self.name == self.db1.last_table_inserted_into, "Last table inserted into is not the same as the table being queried"
+        return self.select("SELECT LAST_INSERT_ID()")[0][0]
+
     def get_nullable_columns(self):
-        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
+        information_schema_table = MysqlTable(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
         query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS where TABLE_SCHEMA = '{self.db1.DB_DATABASE}' and TABLE_NAME = '{self.name}' and IS_NULLABLE = 'YES'"
         nullable_columns = information_schema_table.select(query)
         return (nullable_columns)
@@ -713,28 +726,29 @@ class MysqlTable(AbstractTable):
         Returns the number of records in table
         """
 
-        num_of_records = self.select(f"SELECT COUNT(*) FROM {self.name};")
+        num_of_records = self.select(f"SELECT COUNT(*) FROM `{self.name}`;")
 
         return num_of_records[0][0]
 
-    def drop(self):
-        query = "DROP TABLE " + self.name + ";"
-        print(query)
+    def drop(self, debug_mode = False):
+        query = "DROP TABLE `" + self.name + "`;"
+        if debug_mode:
+            print(query)
         self.db1.execute(query)
 
     # @save_migration #TODO: Uncomment
-    def create(self, foreign_keys=None):
+    def create(self, foreign_keys=None, debug_mode = False):
         assert len(self.columns) == len(self.types)
         assert self.columns[0] == self.id_column_name
         assert self.types[0].lower() == "int"
 
         column_type_pairs = list(zip(self.columns, self.types))[1:]
         fields = ", ".join(
-            [f"{column} {type_.upper()}" for column, type_ in column_type_pairs]
+            [f"`{column}` {type_.upper()}" for column, type_ in column_type_pairs]
         )
-        query = f"CREATE TABLE {self.name} ({self.id_column_name} INT UNSIGNED AUTO_INCREMENT PRIMARY KEY, {fields})"
-
-        print(query)
+        query = f"CREATE TABLE `{self.name}` ({self.id_column_name} INT UNSIGNED AUTO_INCREMENT PRIMARY KEY, {fields})"
+        if debug_mode:
+            print(query)
         try:
             self.db1.execute(query)
         except Exception as e:
@@ -747,14 +761,11 @@ class MysqlTable(AbstractTable):
         total_output=[]
         for k in range(len(rows)):
             if k % batch == 0:
-                query = "INSERT INTO " + self.name + " ("
+                query = "INSERT INTO `" + self.name + "` ("
                 for i in range(start_index, len(self.columns)):
                     if i < len(rows[k]) + 1:
-                        # column name containing space needs to be wrapped in `...`, otherwise causes syntax error
-                        if " " in self.columns[i]:
-                            column_name = '`' + self.columns[i] + '`'
-                        else:
-                            column_name = self.columns[i]
+                        # column name containing space/reserved keyword needs to be wrapped in `...`, otherwise causes syntax error
+                        column_name = '`' + self.columns[i] + '`'
                         query += column_name + ","
                 if len(rows) < len(self.columns):
                     print(len(self.columns) - len(rows), "columns were not specified")
@@ -834,17 +845,17 @@ class MysqlTable(AbstractTable):
     def add_foreign_key(self, foreign_key):
         parent_id = foreign_key['parent_id']
         parent = foreign_key['parent']
-        query = "ALTER TABLE " + self.name + " MODIFY " + parent_id + " INT UNSIGNED"
+        query = "ALTER TABLE `" + self.name + "` MODIFY " + parent_id + " INT UNSIGNED"
         print(query)
         self.db1.execute(query)
-        query = "ALTER TABLE " + self.name + " ADD FOREIGN KEY (" + parent_id + ") REFERENCES " + parent + "("+self.id_column_name+")"
+        query = "ALTER TABLE `" + self.name + "` ADD FOREIGN KEY (" + parent_id + ") REFERENCES " + parent + "("+self.id_column_name+")"
         print(query)
         self.db1.execute(query)
 
     @save_migration
     def add_column(self, column_name, column_type):
         assert len(column_name) > 1
-        command = "ALTER TABLE " + self.name + " ADD COLUMN " + column_name + " " + column_type
+        command = "ALTER TABLE `" + self.name + "` ADD COLUMN `" + column_name + "` " + column_type
         try:
             self.db1.execute(command)
             self.columns.append(column_name)
@@ -855,7 +866,7 @@ class MysqlTable(AbstractTable):
     @save_migration
     def drop_column(self, column_name):
         assert len(column_name) > 1
-        command = "ALTER TABLE " + self.name + " DROP COLUMN " + column_name
+        command = "ALTER TABLE `" + self.name + "` DROP COLUMN " + column_name
         try:
             print(command)
             self.db1.execute(command)
@@ -869,7 +880,7 @@ class MysqlTable(AbstractTable):
     @save_migration
     def modify_column(self, column_name, new_column_type):
         assert len(column_name) > 1
-        command = "ALTER TABLE " + self.name + " MODIFY COLUMN " + column_name + " " + new_column_type
+        command = "ALTER TABLE `" + self.name + "` MODIFY COLUMN `" + column_name + "` " + new_column_type
         print(command)
         try:
             self.db1.execute(command)
@@ -882,9 +893,10 @@ class MysqlTable(AbstractTable):
 ############### XLSX ##################
 
 class XlsxTable(AbstractTable):
-    def __init__(self, db1, name, columns=None, types=None, id_column_name = "id"):
+    def __init__(self, db1, name, columns=None, types=None, id_column_name = "id", number_of_retries=5):
         super().__init__(db1, name, columns, types)
         self.id_column_name = id_column_name
+        self.NUMBER_OF_RETRIES = number_of_retries
 
         table_filename = f"{self.name}.csv" if self.db1.is_csv else f"{self.name}.xlsx"
         self.table_directory_path: pathlib.Path = self.db1.db_directory_path / table_filename
@@ -948,23 +960,32 @@ class XlsxTable(AbstractTable):
             column for column, type_ in self.column_type_dict.items() if type_ == "datetime"
         ]
 
-        try:
-            if self.db1.is_csv:
-                df = pd.read_csv(
-                    self.table_directory_path, dtype=column_type_map,
-                    parse_dates=date_columns, encoding='utf-8'
-                )
-            else:
-                df = pd.read_excel(
-                    self.table_directory_path, dtype=column_type_map, 
-                    parse_dates=date_columns
-                )
-
-            df.replace({np.nan: None}, inplace=True)
-        except Exception as e:
-            print(f"Error while reading data into XlsxTable: {e}")
-            df = pd.DataFrame(columns=self.columns)
+        # BUG: If XlsxTable is being accessed by multiple threads, read operation
+        # might fail due to race conditions. Add retry mechanism to handle these cases.
+        for attempt in range(self.NUMBER_OF_RETRIES):
+            try:
+                df = self._select(column_type_map, date_columns)
+            except Exception:
+                # print(f"Error while reading data into XlsxTable: {e}")
+                # df = pd.DataFrame(columns=self.columns)
+                if attempt < self.NUMBER_OF_RETRIES - 1:
+                    time.sleep(0.1)
+                else:
+                    print(f"Failed to read data from {self.table_directory_path}, returning empty DataFrame")
+                    df = pd.DataFrame(columns=self.columns)
+        return df
 
+    def _select(self, column_type_map, date_columns):
+        if self.db1.is_csv:
+            df = pd.read_csv(
+                self.table_directory_path, dtype=column_type_map, parse_dates=date_columns,
+                encoding='utf-8'
+            )
+        else:
+            df = pd.read_excel(
+                self.table_directory_path, dtype=column_type_map, parse_dates=date_columns
+            )
+        df.replace({np.nan: None}, inplace=True)
         return df
 
     def insert_from_df(self, df, batch=1, try_mode=False, debug_mode=False, adjust_df=False, insert_id=False):
@@ -1014,6 +1035,7 @@ class XlsxTable(AbstractTable):
         df.reset_index(drop=True, inplace=True)
 
         self._save_table(df)
+        self.db1.last_table_inserted_into = self.name
 
     def replace_from_df(self, df):
         assert len(df.columns) == len(self.columns)  # +1 because of id column
@@ -1113,4 +1135,9 @@ class XlsxTable(AbstractTable):
             df.drop(df[df[where_variable] == where_value].index, inplace=True)
             deleted_count = number_of_records - len(df)
         self.replace_from_df(df)
-        return deleted_count
+        return deleted_count    
+
+    def extract_last_id(self) -> Any:
+        assert self.name == self.db1.last_table_inserted_into, "Last table inserted into is not the same as the table being queried"
+        df = self.select_to_df()
+        return df.iloc[-1][self.id_column_name]
